from abc import ABC, abstractmethod
from typing import Dict

from sd_metrics_lib.calculators.metrics import MetricCalculator
from sd_metrics_lib.utils.enums import VelocityTimeUnit
from sd_metrics_lib.utils import time as timeutil
from sd_metrics_lib.sources.story_points import StoryPointExtractor
from sd_metrics_lib.sources.tasks import TaskProvider
from sd_metrics_lib.sources.worklog import WorklogExtractor, TaskTotalSpentTimeExtractor


class AbstractMetricCalculator(MetricCalculator, ABC):

    def __init__(self) -> None:
        self.data_fetched = False

    def calculate(self, velocity_time_unit=VelocityTimeUnit.DAY) -> Dict[str, float]:
        if not self.is_data_fetched():
            self._extract_data_from_tasks()
            self.mark_data_fetched()
        self._calculate_metric(velocity_time_unit)
        return self.get_metric()

    def mark_data_fetched(self):
        self.data_fetched = True

    def is_data_fetched(self):
        return self.data_fetched is True

    @abstractmethod
    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        pass

    @abstractmethod
    def _extract_data_from_tasks(self):
        pass

    @abstractmethod
    def get_metric(self):
        pass


class UserVelocityCalculator(AbstractMetricCalculator):

    def __init__(self, task_provider: TaskProvider,
                 story_point_extractor: StoryPointExtractor,
                 worklog_extractor: WorklogExtractor) -> None:
        super().__init__()
        self.task_provider = task_provider
        self.story_point_extractor = story_point_extractor
        self.worklog_extractor = worklog_extractor

        self.velocity_per_user = {}
        self.resolved_story_points_per_user = {}
        self.time_in_seconds_spent_per_user = {}

    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        for user in self.resolved_story_points_per_user:
            spent_time_in_seconds = self.time_in_seconds_spent_per_user[user]
            if spent_time_in_seconds != 0:
                spent_time = timeutil.convert_time(spent_time_in_seconds, time_unit)
                developer_velocity = self.resolved_story_points_per_user[user] / spent_time
                if developer_velocity != 0:
                    self.velocity_per_user[user] = developer_velocity

    def _extract_data_from_tasks(self):
        tasks = self.task_provider.get_tasks()
        for task in tasks:
            task_story_points = self.story_point_extractor.get_story_points(task)
            if task_story_points is not None and task_story_points > 0:
                time_user_worked_on_task = self.worklog_extractor.get_work_time_per_user(task)

                self._sum_story_points_and_worklog(task_story_points, time_user_worked_on_task)

    def get_metric(self):
        return self.velocity_per_user

    def get_story_points(self):
        return self.resolved_story_points_per_user

    def get_spent_time(self):
        return self.time_in_seconds_spent_per_user

    def _sum_story_points_and_worklog(self, task_story_points, time_user_worked_on_task):
        task_total_spent_time = float(sum(time_user_worked_on_task.values()))
        if task_total_spent_time == 0:
            return

        for user in time_user_worked_on_task.keys():
            if user not in self.resolved_story_points_per_user:
                self.resolved_story_points_per_user[user] = 0.
            if user not in self.time_in_seconds_spent_per_user:
                self.time_in_seconds_spent_per_user[user] = 0

        for user in time_user_worked_on_task.keys():
            story_point_ratio = time_user_worked_on_task[user] / task_total_spent_time
            self.resolved_story_points_per_user[user] += task_story_points * story_point_ratio
            self.time_in_seconds_spent_per_user[user] += time_user_worked_on_task[user]


class GeneralizedTeamVelocityCalculator(AbstractMetricCalculator):

    def __init__(self, task_provider: TaskProvider,
                 story_point_extractor: StoryPointExtractor,
                 time_extractor: TaskTotalSpentTimeExtractor) -> None:
        super().__init__()
        self.total_resolved_story_points = 0
        self.total_spent_time_in_seconds = 0
        self.velocity = None

        self.task_provider = task_provider
        self.story_point_extractor = story_point_extractor
        self.time_extractor = time_extractor

    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        spent_time = timeutil.convert_time(self.total_spent_time_in_seconds, time_unit)
        story_points = self.total_resolved_story_points

        if spent_time == 0:
            self.velocity = 0
        else:
            self.velocity = story_points / spent_time

    def _extract_data_from_tasks(self):
        tasks = self.task_provider.get_tasks()
        for task in tasks:
            task_story_points = self.story_point_extractor.get_story_points(task)
            if task_story_points is not None and task_story_points > 0:
                time_spent_on_task = self.time_extractor.get_total_spent_time(task)

                self._sum_story_points_and_worklog(task_story_points, time_spent_on_task)

    def get_metric(self):
        return self.velocity

    def get_story_points(self):
        return self.total_resolved_story_points

    def get_spent_time(self):
        return self.total_spent_time_in_seconds

    def _sum_story_points_and_worklog(self, task_story_points: float, task_total_spent_time: int):
        if task_total_spent_time == 0:
            return

        self.total_resolved_story_points += task_story_points
        self.total_spent_time_in_seconds += task_total_spent_time



import unittest
from typing import Optional, List

from sd_metrics_lib.sources.tasks import TaskProvider, CachingTaskProvider
from sd_metrics_lib.utils.cache import CacheProtocol, CacheKeyBuilder


class CountingProvider(TaskProvider):
    def __init__(self, tasks: list, query: Optional[str] = None, additional_fields: Optional[List[str]] = None):
        self._tasks = tasks
        self.calls = 0
        self.query = query
        self.additional_fields = additional_fields

    def get_tasks(self) -> list:
        self.calls += 1
        return list(self._tasks)


class FakeDjangoCache(CacheProtocol):
    def __init__(self):
        self._store = {}

    def get(self, key: str):
        return self._store.get(key, None)

    def set(self, key: str, value):
        self._store[key] = value


class CachingTaskProviderTestCase(unittest.TestCase):

    def test_should_use_cache_on_exact_hit(self):
        # given
        cache = {}
        provider = CountingProvider(tasks=[{"id": 1}], query="project = X", additional_fields=["a", "b"])
        caching = CachingTaskProvider(provider, cache)

        # when
        result1 = caching.get_tasks()
        result2 = caching.get_tasks()

        # then
        self.assertEqual(result1, [{"id": 1}])
        self.assertEqual(result2, [{"id": 1}])
        self.assertEqual(provider.calls, 1)

    def test_should_reuse_superset_when_exact_missing(self):
        # given
        cache = {}
        superset_provider = CountingProvider(tasks=[{"id": 1, "a": 1, "b": 2}], query="Q", additional_fields=["a", "b"])
        CachingTaskProvider(superset_provider, cache).get_tasks()
        self.assertEqual(superset_provider.calls, 1)

        # when
        subset_provider = CountingProvider(tasks=[{"id": 999}], query="Q", additional_fields=["a"])  # would be used only if miss
        result = CachingTaskProvider(subset_provider, cache).get_tasks()

        # then
        self.assertEqual(result, [{"id": 1, "a": 1, "b": 2}])
        self.assertEqual(subset_provider.calls, 0)

    def test_should_normalize_fields_to_avoid_duplicates(self):
        # given
        cache = {}
        provider1 = CountingProvider(tasks=[1, 2, 3], query="Q2", additional_fields=["b", "a", "a"])  # unsorted + dup
        caching1 = CachingTaskProvider(provider1, cache)
        caching1.get_tasks()  # warm cache
        provider2 = CountingProvider(tasks=[9], query="Q2", additional_fields=["a", "b"])  # would be used on miss
        caching2 = CachingTaskProvider(provider2, cache)

        # when
        result2 = caching2.get_tasks()

        # then
        self.assertEqual(result2, [1, 2, 3])
        self.assertEqual(provider2.calls, 0)
        self.assertEqual(provider1.calls, 1)

    def test_should_hit_superset_even_if_metadata_unsorted(self):
        # given
        cache = {}
        superset_provider = CountingProvider(tasks=[{"id": 7, "a": 10, "b": 20}], query="Q3", additional_fields=["a", "b"])
        CachingTaskProvider(superset_provider, cache).get_tasks()
        self.assertEqual(superset_provider.calls, 1)
        partial = CacheKeyBuilder.create_query_only_key_partial("Q3")
        meta_key = CacheKeyBuilder.create_meta_data_key(partial)
        cache[meta_key] = [["b", "a"]]

        # when
        subset_provider = CountingProvider(tasks=[{"id": 999}], query="Q3", additional_fields=["a"])  # used only on miss
        result = CachingTaskProvider(subset_provider, cache).get_tasks()

        # then
        self.assertEqual(result, [{"id": 7, "a": 10, "b": 20}])
        self.assertEqual(subset_provider.calls, 0)

    def test_should_work_with_none_query_and_empty_fields(self):
        # given
        cache = {}
        provider1 = CountingProvider(tasks=["x"], query=None, additional_fields=None)
        caching1 = CachingTaskProvider(provider1, cache)
        caching1.get_tasks()  # warm cache
        provider2 = CountingProvider(tasks=["y"], query=None, additional_fields=None)
        caching2 = CachingTaskProvider(provider2, cache)

        # when
        result2 = caching2.get_tasks()

        # then
        self.assertEqual(result2, ["x"])  # cached value
        self.assertEqual(provider2.calls, 0)
        self.assertEqual(provider1.calls, 1)

    def test_should_support_django_style_cache(self):
        # given
        django_cache = FakeDjangoCache()
        provider = CountingProvider(tasks=[{"id": 42}], query="D", additional_fields=["f1"])
        caching = CachingTaskProvider(provider, django_cache)
        caching.get_tasks()  # warm cache through adapter
        provider2 = CountingProvider(tasks=[{"id": 999}], query="D", additional_fields=["f1"])  # ignored if hit
        caching2 = CachingTaskProvider(provider2, django_cache)

        # when
        result2 = caching2.get_tasks()

        # then
        self.assertEqual(result2, [{"id": 42}])
        self.assertEqual(provider2.calls, 0)
        self.assertEqual(provider.calls, 1)


if __name__ == "__main__":
    unittest.main()



import concurrent
import math
from concurrent.futures import ThreadPoolExecutor, wait
from typing import Iterable

from sd_metrics_lib.sources.tasks import TaskProvider


class JiraTaskProvider(TaskProvider):

    def __init__(self,
                 jira_client,
                 query: str,
                 additional_fields: Iterable[str] = None,
                 thread_pool_executor: ThreadPoolExecutor = None) -> None:
        self.jira_client = jira_client
        self.query = query.strip()
        self.additional_fields = additional_fields
        if additional_fields is None:
            self._expand_str = None
        else:
            # For Jira, additional_fields correspond to expand values (e.g., 'changelog')
            self._expand_str = ",".join(self.additional_fields)
        self.thread_pool_executor = thread_pool_executor

    def get_tasks(self):
        tasks = self._fetch_tasks(self.query, self._expand_str)
        if self.additional_fields and 'subtasks' in self.additional_fields:
            self._fetch_child_tasks_and_replace_subtasks_field(tasks)

        return tasks

    def _fetch_tasks(self, query: str, expand_str: str):
        first_page = self.jira_client.jql(query, expand=expand_str, limit=self._get_task_fetch_amount())
        first_page_tasks = first_page.get("issues", [])
        tasks_total_count = first_page.get("total", len(first_page_tasks))
        page_len = len(first_page_tasks)
        if tasks_total_count == 0 or page_len == 0:
            return []

        tasks = []
        tasks.extend(first_page_tasks)
        if page_len < tasks_total_count:
            amount_of_fetches = math.ceil(tasks_total_count / float(page_len))

            if self.thread_pool_executor is None:
                self._fetch_task_sync(tasks, amount_of_fetches, page_len)
            else:
                self._fetch_task_concurrently(tasks, amount_of_fetches, page_len)
        return tasks

    def _fetch_task_concurrently(self, tasks, amount_of_fetches, page_len):
        features = []
        for i in range(1, amount_of_fetches):
            next_search_start = i * page_len
            feature = self.thread_pool_executor.submit(self.jira_client.jql,
                                                       self.query,
                                                       expand=self._expand_str,
                                                       limit=self._get_task_fetch_amount(),
                                                       start=next_search_start)
            features.append(feature)
        done, not_done = wait(features, return_when=concurrent.futures.ALL_COMPLETED)
        for feature in done:
            tasks.extend(feature.result().get("issues", []))

    def _fetch_task_sync(self, tasks, amount_of_fetches, page_len):
        for i in range(1, amount_of_fetches):
            start = i * page_len
            current_page_result = self.jira_client.jql(self.query,
                                                       expand=self._expand_str,
                                                       limit=self._get_task_fetch_amount(),
                                                       start=start)
            current_page_tasks = current_page_result.get("issues", [])
            tasks.extend(current_page_tasks)

    def _fetch_child_tasks_and_replace_subtasks_field(self, jira_tasks: Iterable[dict]):
        if not jira_tasks:
            return

        child_tasks_ids = []
        task_id_to_child_tasks_ids = {}
        for jira_task in jira_tasks:
            subtasks = jira_task.get('fields', {}).get('subtasks', [])
            if subtasks:
                subtasks_ids = [subtask.get('key') for subtask in subtasks if subtask.get('key')]
                if subtasks_ids:
                    child_tasks_ids.extend(subtasks_ids)
                    task_id_to_child_tasks_ids[jira_task['key']] = subtasks_ids

        if not child_tasks_ids:
            return

        child_task_id_to_child_task = self._fetch_tasks_by_id(child_tasks_ids)
        for task in jira_tasks:
            task_key = task['key']
            if task_key in task_id_to_child_tasks_ids:
                task['fields']['subtasks'] = self._create_child_task_list(
                    task_key,
                    task_id_to_child_tasks_ids,
                    child_task_id_to_child_task
                )

    def _fetch_tasks_by_id(self, task_ids):
        query = "key in (" + ", ".join(task_ids) + ")"
        child_tasks = self._fetch_tasks(
            query,
            ",".join([field for field in self.additional_fields if field != 'subtasks'])
        )
        return {task['key']: task for task in child_tasks}

    @staticmethod
    def _create_child_task_list(task_key, task_to_child_tasks_ids, child_task_id_to_child_task):
        return [
            child_task_id_to_child_task[child_key]
            for child_key in task_to_child_tasks_ids[task_key]
            if child_key in child_task_id_to_child_task
        ]

    def _get_task_fetch_amount(self):
        if self.thread_pool_executor is None:
            return 100
        else:
            return 50



from abc import abstractmethod, ABC
from typing import Optional, List, Tuple, Set, Iterable
from typing import Union

from sd_metrics_lib.utils.cache import (
    DictProtocol,
    CacheProtocol,
    DictToCacheProtocolAdapter,
    CacheKeyBuilder,
    SupersetResolver
)


class TaskProvider(ABC):

    @abstractmethod
    def get_tasks(self) -> list:
        pass


class ProxyTaskProvider(TaskProvider):

    def __init__(self, tasks: list) -> None:
        self.tasks = tasks

    def get_tasks(self) -> list:
        return self.tasks


class CachingTaskProvider(TaskProvider):

    def __init__(self, provider: TaskProvider,
                 cache: Optional[Union[DictProtocol, CacheProtocol]] = None) -> None:
        if cache is not None and isinstance(cache, DictProtocol):
            self.cache: Optional[CacheProtocol] = DictToCacheProtocolAdapter(cache)
        else:
            self.cache = cache  # type: ignore[assignment]

        self.provider = provider
        self.query = getattr(provider, 'query', None)
        self.additional_fields = getattr(provider, 'additional_fields', None)

    def get_tasks(self):
        cached = self._try_fetch_from_cache()
        if cached is not None:
            return cached
        tasks = self.provider.get_tasks()
        self._store_in_cache(tasks)
        return tasks

    def _try_fetch_from_cache(self):
        if self.cache is None:
            return None

        exact = self._fetch_exact_cache_hit()
        if exact is not None:
            return exact

        superset = self._fetch_superset_cache_hit()
        if superset is not None:
            return superset

        return None

    def _store_in_cache(self, tasks):
        if self.cache is None:
            return
        normalized_fields = CacheKeyBuilder.normalize_fields(self.additional_fields)
        self._store_tasks_under_data_key(tasks, normalized_fields)
        self._ensure_fieldset_list_updated(normalized_fields)

    def _fetch_exact_cache_hit(self):
        partial_key = CacheKeyBuilder.create_query_only_key_partial(self.query)
        data_key = CacheKeyBuilder.create_full_data_key(partial_key,
                                                        CacheKeyBuilder.normalize_fields(self.additional_fields))
        hit = self.cache.get(data_key)  # type: ignore[union-attr]
        if hit is not None:
            return hit
        return None

    def _fetch_superset_cache_hit(self):
        requested_fields = CacheKeyBuilder.normalize_fields(self.additional_fields)
        meta_key = CacheKeyBuilder.create_meta_data_key(CacheKeyBuilder.create_query_only_key_partial(self.query))
        available_fieldsets = self._load_cached_fieldsets(meta_key)
        compatible_available_fieldset = SupersetResolver.find_superset_fieldset(requested_fields, available_fieldsets)
        if compatible_available_fieldset is not None:
            superset_data_key = CacheKeyBuilder.create_full_data_key(
                CacheKeyBuilder.create_query_only_key_partial(self.query),
                compatible_available_fieldset)
            superset_value = self.cache.get(superset_data_key)  # type: ignore[union-attr]
            if superset_value is not None:
                return superset_value
        return None

    def _store_tasks_under_data_key(self, tasks, fields: Iterable[str]):
        partial_key = CacheKeyBuilder.create_query_only_key_partial(self.query)
        data_key = CacheKeyBuilder.create_full_data_key(partial_key, fields)
        self.cache.set(data_key, tasks)  # type: ignore[union-attr]

    def _ensure_fieldset_list_updated(self, fields: Iterable[str]):
        meta_key = self._create_meta_key_for_query()
        fieldsets = self._load_cached_fieldsets(meta_key)
        fieldset_tuple = tuple(fields)
        if fieldset_tuple not in fieldsets:
            fieldsets.add(fieldset_tuple)
            self._save_cached_fieldsets(meta_key, fieldsets)

    def _create_meta_key_for_query(self) -> str:
        return CacheKeyBuilder.create_meta_data_key(CacheKeyBuilder.create_query_only_key_partial(self.query))

    def _load_cached_fieldsets(self, meta_key: str) -> Set[Tuple[str, ...]]:
        raw = self.cache.get(meta_key)  # type: ignore[union-attr]
        if raw is None:
            return set()
        result: Set[Tuple[str, ...]] = set()
        try:
            for fieldset in raw:
                result.add(tuple(fieldset))
        except Exception:
            return set()
        return result

    def _save_cached_fieldsets(self, meta_key: str, fieldsets: Set[Tuple[str, ...]]):
        serializable: List[List[str]] = [list(t) for t in fieldsets]
        self.cache.set(meta_key, serializable)  # type: ignore[union-attr]



import base64
from typing import Protocol, runtime_checkable, Any, Iterable, List, Tuple, Optional


@runtime_checkable
class CacheProtocol(Protocol):
    def get(self, key: str) -> Any: ...

    def set(self, key: str, value: Any) -> None: ...


@runtime_checkable
class DictProtocol(Protocol):
    def __getitem__(self, key: str) -> Any: ...

    def __setitem__(self, key: str, value: Any) -> None: ...


class DictToCacheProtocolAdapter(CacheProtocol):
    def __init__(self, mapping: DictProtocol) -> None:
        self._dict = mapping

    def get(self, key: str) -> Any:
        try:
            return self._dict[key]
        except KeyError:
            return None

    def set(self, key: str, value: Any) -> None:
        self._dict[key] = value


class CacheKeyBuilder:
    DATA_PREFIX = "data||"
    META_PREFIX = "meta||"

    @staticmethod
    def normalize_fields(fields: Optional[Iterable[str]]) -> List[str]:
        if not fields:
            return []
        return sorted(set(fields))

    @staticmethod
    def create_query_only_key_partial(query: Optional[str]) -> str:
        if query is None:
            return "none_query||"
        return base64.b64encode(query.encode("ascii")).decode("ascii") + "||"

    @staticmethod
    def create_full_data_key(query_only_key_partial: str, fields: Optional[Iterable[str]]) -> str:
        if not fields:
            return f"{CacheKeyBuilder.DATA_PREFIX}{query_only_key_partial}"
        normalized = CacheKeyBuilder.normalize_fields(fields)
        return f"{CacheKeyBuilder.DATA_PREFIX}{query_only_key_partial}" + "_".join(normalized)

    @staticmethod
    def create_meta_data_key(query_only_key_partial: str) -> str:
        return f"{CacheKeyBuilder.META_PREFIX}{query_only_key_partial}fieldsets"


class SupersetResolver:
    @staticmethod
    def find_superset_fieldset(requested_fields: Iterable[str], available_fieldsets: Iterable[Tuple[str, ...]]) -> \
            Optional[Tuple[str, ...]]:
        requested_set = set(requested_fields)
        for fieldset in available_fieldsets:
            if requested_set.issubset(fieldset):
                return fieldset
        return None



import unittest

from sd_metrics_lib.calculators.velocity import UserVelocityCalculator
from sd_metrics_lib.sources.jira.story_points import JiraCustomFieldStoryPointExtractor
from sd_metrics_lib.sources.jira.worklog import JiraStatusChangeWorklogExtractor
from sd_metrics_lib.sources.tasks import ProxyTaskProvider

TEST_USER = 'test_user'


class VelocityCalculatorTestCase(unittest.TestCase):

    def test_should_not_fail_on_empty_jira_task(self):
        # given
        empty_task = {}
        task_provider = ProxyTaskProvider([empty_task])
        t_shirt_story_point_extractor = JiraCustomFieldStoryPointExtractor('customfield_00000')
        jira_worklog_extractor = JiraStatusChangeWorklogExtractor(['12207'],
                                                                  time_format='%Y-%m-%dT%H:%M:%S.%f%z')
        # when
        self.velocity_calculator = UserVelocityCalculator(task_provider=task_provider,
                                                          story_point_extractor=t_shirt_story_point_extractor,
                                                          worklog_extractor=jira_worklog_extractor)

        velocity = self.velocity_calculator.calculate()

        # then
        self.assertEqual(0, len(velocity.keys()), 'Empty task must not provide any velocity data')

    def test_should_calculate_velocity_for_one_worklog(self):
        # given
        histories = self.__create_history_entries_with_status_change()
        task = self.__create_task_with_log_data(histories)

        task_provider = ProxyTaskProvider([task])
        t_shirt_story_point_extractor = JiraCustomFieldStoryPointExtractor('customfield_00000')
        jira_worklog_extractor = JiraStatusChangeWorklogExtractor(['12207'],
                                                                  time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                                                                  use_status_codes=True)

        # when
        self.velocity_calculator = UserVelocityCalculator(task_provider=task_provider,
                                                          story_point_extractor=t_shirt_story_point_extractor,
                                                          worklog_extractor=jira_worklog_extractor)

        velocity = self.velocity_calculator.calculate()

        # then
        self.assertEqual(1, len(velocity.keys()), 'Missing calculated velocity data')
        self.assertEqual(24, velocity[TEST_USER], 'Must be calculated velocity for test user')

    def test_should_calculate_velocity_for_few_worklogs(self):
        # given
        histories = self.__create_history_entries_with_status_change()
        histories.extend(self.__create_history_entries_with_status_change())
        task = self.__create_task_with_log_data(histories)

        task_provider = ProxyTaskProvider([task])
        t_shirt_story_point_extractor = JiraCustomFieldStoryPointExtractor('customfield_00000')
        jira_worklog_extractor = JiraStatusChangeWorklogExtractor(['12207'],
                                                                  time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                                                                  use_status_codes=True)

        # when
        self.velocity_calculator = UserVelocityCalculator(task_provider=task_provider,
                                                          story_point_extractor=t_shirt_story_point_extractor,
                                                          worklog_extractor=jira_worklog_extractor)

        velocity = self.velocity_calculator.calculate()

        # then
        self.assertEqual(1, len(velocity.keys()), 'Missing calculated velocity data')
        self.assertEqual(12, velocity[TEST_USER], 'Must be calculated velocity for test user')

    def test_should_calculate_velocity_for_few_tasks(self):
        # given
        task = self.__create_task_with_log_data(self.__create_history_entries_with_status_change())
        task2 = self.__create_task_with_log_data(self.__create_history_entries_with_status_change(), story_points=2)

        task_provider = ProxyTaskProvider([task, task2])
        t_shirt_story_point_extractor = JiraCustomFieldStoryPointExtractor('customfield_00000')
        jira_worklog_extractor = JiraStatusChangeWorklogExtractor(['12207'],
                                                                  time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                                                                  use_status_codes=True)

        # when
        self.velocity_calculator = UserVelocityCalculator(task_provider=task_provider,
                                                          story_point_extractor=t_shirt_story_point_extractor,
                                                          worklog_extractor=jira_worklog_extractor)

        velocity = self.velocity_calculator.calculate()

        # then
        self.assertEqual(1, len(velocity.keys()), 'Missing calculated velocity data')
        self.assertEqual(20, velocity[TEST_USER], 'Must be calculated velocity for test user')

    @staticmethod
    def __create_task_with_log_data(histories, story_points=3):
        task = {}
        task['fields'] = {}
        task['fields']['customfield_00000'] = story_points
        task['changelog'] = {}
        task['changelog']['histories'] = histories
        return task

    def __create_history_entries_with_status_change(self):
        end_date_history_entry_item = self.__create_status_change_history_entry_item(to_status='1',
                                                                                     from_status='12207')
        history_entry_end_date = {}
        history_entry_end_date['items'] = [end_date_history_entry_item]
        history_entry_end_date['created'] = '2022-02-01T11:00:00.000-0500'

        start_date_history_entry_item = self.__create_status_change_history_entry_item(to_status='12207',
                                                                                       from_status='1')
        assignee_change_history_entry_item = self.__create_assignee_change_history_entry_item()
        history_entry_start_date = {}
        history_entry_start_date['items'] = [start_date_history_entry_item, assignee_change_history_entry_item]
        history_entry_start_date['created'] = '2022-02-01T10:00:00.000-0500'
        histories = [history_entry_end_date, history_entry_start_date]
        return histories

    @staticmethod
    def __create_status_change_history_entry_item(to_status, from_status):
        return {'items': {}, 'fieldId': 'status', 'to': to_status, 'from': from_status}

    @staticmethod
    def __create_assignee_change_history_entry_item():
        return {'fieldId': 'assignee', 'to': TEST_USER, 'toString': TEST_USER}


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from typing import Optional

from sd_metrics_lib.sources.worklog import WorklogExtractor
from sd_metrics_lib.sources.abstract_worklog import AbstractStatusChangeWorklogExtractor
from sd_metrics_lib.sources.worklog import TaskTotalSpentTimeExtractor
from sd_metrics_lib.utils.worktime import SimpleWorkTimeExtractor, WorkTimeExtractor


class JiraWorklogExtractor(WorklogExtractor):

    def __init__(self, jira_client, user_filter: list[str] = None, include_subtask_worklog=False) -> None:
        self.jira_client = jira_client
        self.user_filter = user_filter
        self.include_subtask_worklog = include_subtask_worklog

    def get_work_time_per_user(self, task):
        worklogs = self._get_worklog_for_task_with_subtasks(task)

        working_time_per_user = {}
        for worklog in worklogs:
            worklog_user = self._extract_user_from_worklog(worklog)
            if self._is_allowed_user(worklog_user):
                worklog_time_spent = self._extract_time_in_seconds_from_worklog(worklog)

                working_time_per_user[worklog_user] = working_time_per_user.get(worklog_user, 0) + worklog_time_spent

        return working_time_per_user

    def _get_worklog_for_task_with_subtasks(self, task):
        worklogs = []
        worklogs.extend(self._get_worklogs_from_jira(task['key']))
        if self.include_subtask_worklog:
            try:
                sub_task_keys = [subtask["key"] for subtask in task["fields"]["subtasks"]]
                for subtask in sub_task_keys:
                    worklogs.extend(self._get_worklogs_from_jira(subtask))
            except AttributeError:
                pass
        return worklogs

    def _get_worklogs_from_jira(self, task_key: str):
        data = self.jira_client.issue_get_worklog(task_key)
        if 'worklogs' in data:
            return data['worklogs']
        return data

    def _is_allowed_user(self, worklog_user):
        if self.user_filter is None:
            return True
        if worklog_user in self.user_filter:
            return True
        return False

    @staticmethod
    def _extract_time_in_seconds_from_worklog(worklog):
        return worklog["timeSpentSeconds"]

    @staticmethod
    def _extract_user_from_worklog(worklog):
        return worklog["author"]["accountId"]


SIMPLE_WORKTIME_EXTRACTOR = SimpleWorkTimeExtractor()

class JiraStatusChangeWorklogExtractor(AbstractStatusChangeWorklogExtractor):

    def __init__(self, transition_statuses: list[str],
                 user_filter: list[str] = None,
                 time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                 use_user_name=False,
                 use_status_codes=False,
                 worktime_extractor: WorkTimeExtractor = SIMPLE_WORKTIME_EXTRACTOR) -> None:

        super().__init__(transition_statuses=transition_statuses,
                         user_filter=user_filter,
                         worktime_extractor=worktime_extractor)
        self.time_format = time_format
        self.use_user_name = use_user_name
        self.use_status_codes = use_status_codes

    def _extract_chronological_changes_sequence(self, task):
        if not isinstance(task, dict):
            return []
        if 'changelog' not in task or 'histories' not in task['changelog']:
            return []

        changelog_history = []
        changelog = task['changelog']['histories']
        for history_entry in changelog:
            if 'items' not in history_entry:
                continue
            for history_entry_item in history_entry['items']:
                if self._is_status_change_entry(history_entry_item) or self._is_user_change_entry(history_entry_item):
                    history_entry_item['created'] = history_entry['created']
                    history_entry_item['author'] = history_entry.get('author', {})
                    changelog_history.append(history_entry_item)

        changelog_history.reverse()  # Jira returns newest first; reverse to chronological
        return changelog_history

    def _is_user_change_entry(self, changelog_entry):
        return 'fieldId' in changelog_entry and changelog_entry['fieldId'] == 'assignee'

    def _is_status_change_entry(self, changelog_entry):
        return 'fieldId' in changelog_entry and changelog_entry['fieldId'] == 'status'

    def _extract_user_from_change(self, changelog_entry):
        if self.use_user_name:
            return changelog_entry['toString']
        else:
            return changelog_entry['to']

    def _extract_change_time(self, changelog_entry):
        return datetime.strptime(changelog_entry['created'], self.time_format)

    def _is_status_changed_into_required(self, changelog_entry):
        if self.transition_statuses is None:
            return True
        if self.use_status_codes:
            return changelog_entry['to'] in self.transition_statuses
        else:
            return changelog_entry['toString'] in self.transition_statuses

    def _is_status_changed_from_required(self, changelog_entry):
        if self.transition_statuses is None:
            return True
        if self.use_status_codes:
            return changelog_entry['from'] in self.transition_statuses
        else:
            return changelog_entry['fromString'] in self.transition_statuses

    def _is_current_status_a_required_status(self, task):
        if self.transition_statuses is None:
            return True
        if 'fields' not in task or 'status' not in task['fields']:
            return False
        if self.use_status_codes:
            return task['fields']['status']['id'] in self.transition_statuses
        else:
            return task['fields']['status']['name'] in self.transition_statuses

    def _extract_author_from_changelog_entry(self, changelog_entry) -> Optional[str]:
        author = changelog_entry.get('author', {})
        if not author:
            return None

        if self.use_user_name:
            return author.get('displayName')
        else:
            return author.get('accountId')


class JiraResolutionTimeTaskTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, time_format='%Y-%m-%dT%H:%M:%S.%f%z') -> None:
        self.time_format = time_format

    def get_total_spent_time(self, task) -> int:
        resolution_date_str = task['fields']['resolutiondate']
        if resolution_date_str is None:
            return 0

        resolution_date = datetime.strptime(resolution_date_str, self.time_format)
        creation_date = datetime.strptime(task['fields']['created'], self.time_format)
        spent_time = (resolution_date - creation_date)
        return int(spent_time.total_seconds())




from sd_metrics_lib.utils.enums import VelocityTimeUnit

SECONDS_IN_HOUR = 3600
WORKING_HOURS_PER_DAY = 8
WORKING_DAYS_PER_WEEK = 5
WORKING_WEEKS_IN_MONTH = 4

# Python's date.weekday(): Monday=0
WEEKDAY_FRIDAY = 4


def get_seconds_in_day() -> int:
    return WORKING_HOURS_PER_DAY * SECONDS_IN_HOUR


def convert_time(spent_time_in_seconds: int, time_unit: VelocityTimeUnit):
    if spent_time_in_seconds is None:
        return 0
    if time_unit == VelocityTimeUnit.HOUR:
        return spent_time_in_seconds / SECONDS_IN_HOUR
    elif time_unit == VelocityTimeUnit.DAY:
        return spent_time_in_seconds / SECONDS_IN_HOUR / WORKING_HOURS_PER_DAY
    elif time_unit == VelocityTimeUnit.WEEK:
        return spent_time_in_seconds / SECONDS_IN_HOUR / WORKING_HOURS_PER_DAY / WORKING_DAYS_PER_WEEK
    elif time_unit == VelocityTimeUnit.MONTH:
        return spent_time_in_seconds / SECONDS_IN_HOUR / WORKING_HOURS_PER_DAY / WORKING_DAYS_PER_WEEK / WORKING_WEEKS_IN_MONTH



import datetime
from abc import ABC, abstractmethod

from sd_metrics_lib.utils.time import WEEKDAY_FRIDAY, get_seconds_in_day


class WorkTimeExtractor(ABC):

    @abstractmethod
    def extract_time_from_period(self, start_time_period: datetime.date, end_time_period: datetime.date) -> int | None:
        pass


class SimpleWorkTimeExtractor(WorkTimeExtractor):

    def extract_time_from_period(self,
                                 start_time_period: datetime.date,
                                 end_time_period: datetime.date) -> int | None:
        # Use businesstimedelta lib for more precision calculation
        period_delta = end_time_period - start_time_period

        if period_delta.days > 0:
            work_days = self.__count_work_days(start_time_period, end_time_period)
            round_up_period_days = period_delta.days + 1
            return min(work_days, round_up_period_days) * get_seconds_in_day()
        elif period_delta.total_seconds() < 15 * 60:
            return None
        elif period_delta.total_seconds() < get_seconds_in_day():
            return int(period_delta.total_seconds())
        else:
            return get_seconds_in_day()

    @staticmethod
    def __count_work_days(start_date: datetime.date, end_date: datetime.date):
        # if the start date is on a weekend, forward the date to next Monday
        if start_date.weekday() > WEEKDAY_FRIDAY:
            start_date = start_date + datetime.timedelta(days=7 - start_date.weekday())

        # if the end date is on a weekend, rewind the date to the previous Friday
        if end_date.weekday() > WEEKDAY_FRIDAY:
            end_date = end_date - datetime.timedelta(days=end_date.weekday() - WEEKDAY_FRIDAY)

        if start_date > end_date:
            return 0
        # that makes the difference easy, no remainders etc
        diff_days = (end_date - start_date).days + 1
        weeks = int(diff_days / 7)

        remainder = end_date.weekday() - start_date.weekday() + 1
        if remainder != 0 and end_date.weekday() < start_date.weekday():
            remainder = 5 + remainder

        return weeks * 5 + remainder


class BoundarySimpleWorkTimeExtractor(SimpleWorkTimeExtractor):

    def __init__(self,
                 start_time_boundary: datetime.date,
                 end_time_boundary: datetime.date) -> None:
        self.start_time_boundary = start_time_boundary
        self.end_time_boundary = end_time_boundary

    def extract_time_from_period(self, start_time_period: datetime.date, end_time_period: datetime.date) -> int | None:

        if self.start_time_boundary < start_time_period:
            new_start_period = start_time_period
        else:
            new_start_period = self.start_time_boundary

        if self.end_time_boundary > end_time_period:
            new_end_time_period = end_time_period
        else:
            new_end_time_period = self.end_time_boundary

        return super().extract_time_from_period(new_start_period, new_end_time_period)



from datetime import datetime
from typing import Optional

from sd_metrics_lib.sources.abstract_worklog import AbstractStatusChangeWorklogExtractor
from sd_metrics_lib.sources.worklog import TaskTotalSpentTimeExtractor
from sd_metrics_lib.utils.worktime import WorkTimeExtractor, SimpleWorkTimeExtractor


class AzureStatusChangeWorklogExtractor(AbstractStatusChangeWorklogExtractor):

    def __init__(self,
                 transition_statuses: Optional[list[str]] = None,
                 user_filter: Optional[list[str]] = None,
                 time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                 use_user_name: bool = True,
                 worktime_extractor: WorkTimeExtractor = SimpleWorkTimeExtractor()) -> None:
        super().__init__(transition_statuses=transition_statuses,
                         user_filter=user_filter,
                         worktime_extractor=worktime_extractor)
        self.time_format = time_format
        self.use_user_name = use_user_name

    def _extract_chronological_changes_sequence(self, task):
        fields = task.fields
        return fields.get('CustomExpand.WorkItemUpdate', [])

    def _is_user_change_entry(self, changelog_entry) -> bool:
        fields = changelog_entry.fields
        return fields and 'System.AssignedTo' in fields and fields['System.AssignedTo'].new_value is not None

    def _is_status_change_entry(self, changelog_entry) -> bool:
        fields = changelog_entry.fields
        return fields and 'System.State' in fields and fields['System.State'].new_value is not None

    def _extract_user_from_change(self, changelog_entry) -> str:
        assigned_to = changelog_entry.fields['System.AssignedTo'].new_value
        if self.use_user_name:
            return assigned_to.get('displayName', assigned_to.get('id', self._default_assigned_user()))
        else:
            return assigned_to.get('id', self._default_assigned_user())

    def _extract_change_time(self, changelog_entry):
        date_to_use = None

        fields = changelog_entry.fields
        if fields:
            state_change_field = fields.get('Microsoft.VSTS.Common.StateChangeDate')
            if state_change_field and state_change_field.new_value:
                date_to_use = state_change_field.new_value
            else:
                changed_date_field = fields.get('System.ChangedDate')
                if changed_date_field and changed_date_field.new_value:
                    date_to_use = changed_date_field.new_value

        if not date_to_use:
            # Revised date pretty often contains a placeholder value with date 9999-01-01
            date_to_use = changelog_entry.revised_date

        if isinstance(date_to_use, datetime):
            return date_to_use
        else:
            try:
                return datetime.strptime(date_to_use, self.time_format)
            except ValueError:
                # Sometimes Azure API returns time without milliseconds
                return datetime.strptime(date_to_use, '%Y-%m-%dT%H:%M:%S%z')

    def _is_status_changed_into_required(self, changelog_entry) -> bool:
        if self.transition_statuses is None:
            return True
        return changelog_entry.fields['System.State'].new_value in self.transition_statuses

    def _is_status_changed_from_required(self, changelog_entry) -> bool:
        if self.transition_statuses is None:
            return True
        return changelog_entry.fields['System.State'].old_value in self.transition_statuses

    def _is_current_status_a_required_status(self, task) -> bool:
        if self.transition_statuses is None:
            return True
        current_state = task.fields.get('System.State')
        return current_state in self.transition_statuses

    def _extract_author_from_changelog_entry(self, changelog_entry) -> Optional[str]:
        changed_by = changelog_entry.fields['System.ChangedBy'].new_value
        if changed_by:
            if self.use_user_name:
                return changed_by.get('displayName', changed_by.get('id'))
            else:
                return changed_by.get('id')
        return None


class AzureTaskTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, time_format='%Y-%m-%dT%H:%M:%S.%f%z') -> None:
        self.time_format = time_format

    def get_total_spent_time(self, task) -> int:
        resolution_date_str = task.fields['Microsoft.VSTS.Common.ClosedDate']
        if resolution_date_str is None:
            return 0

        resolution_date = self._convert_to_time(resolution_date_str)
        creation_date = self._convert_to_time(task.fields['System.CreatedDate'])
        spent_time = (resolution_date - creation_date)
        return int(spent_time.total_seconds())

    def _convert_to_time(self, date_string: str) -> datetime:
        try:
            return datetime.strptime(date_string, self.time_format)
        except ValueError:
            # Sometimes Azure API returns time without milliseconds
            return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S%z')



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, Iterable, Optional

from sd_metrics_lib.sources.worklog import WorklogExtractor
from sd_metrics_lib.utils.worktime import WorkTimeExtractor, SimpleWorkTimeExtractor


class AbstractStatusChangeWorklogExtractor(WorklogExtractor, ABC):

    def __init__(self,
                 transition_statuses: Optional[list[str]] = None,
                 user_filter: Optional[list[str]] = None,
                 worktime_extractor: WorkTimeExtractor = SimpleWorkTimeExtractor()) -> None:
        self.transition_statuses = transition_statuses
        self.user_filter = user_filter
        self.worktime_extractor = worktime_extractor

        self.interval_start_time: Optional[datetime] = None
        self.interval_end_time: Optional[datetime] = None

    def get_work_time_per_user(self, task) -> Dict[str, int]:
        working_time_per_user: Dict[str, int] = {}

        changelog_history = list(self._extract_chronological_changes_sequence(task))
        if not changelog_history:
            return working_time_per_user

        last_assigned_user = self._default_assigned_user()
        for changelog_entry in changelog_history:
            is_user_change_entry = self._is_user_change_entry(changelog_entry)
            is_status_change_entry = self._is_status_change_entry(changelog_entry)

            if not is_user_change_entry and not is_status_change_entry:
                continue

            if is_user_change_entry and is_status_change_entry:
                assignee = self._extract_user_from_change(changelog_entry)
                if self._is_allowed_user(assignee):
                    last_assigned_user = assignee

                last_assigned_user = self._get_current_assignee_from_changelog_when_last_assigned_is_unknown(
                    changelog_entry,
                    last_assigned_user
                )
                self._update_time_intervals_and_sum_worklog(changelog_entry, working_time_per_user, last_assigned_user)
            elif is_user_change_entry:
                assignee = self._extract_user_from_change(changelog_entry)
                if self._is_allowed_user(assignee):
                    last_assigned_user = assignee
            elif is_status_change_entry:
                last_assigned_user = self._get_current_assignee_from_changelog_when_last_assigned_is_unknown(
                    changelog_entry,
                    last_assigned_user
                )
                self._update_time_intervals_and_sum_worklog(changelog_entry, working_time_per_user, last_assigned_user)

        if self._is_current_status_a_required_status(task):
            self.interval_end_time = self._now()
            self._sum_working_time(working_time_per_user, last_assigned_user)

        return working_time_per_user

    def _update_time_intervals_and_sum_worklog(self, changelog_entry, working_time_per_user, last_assigned_user):
        change_time = self._extract_change_time(changelog_entry)
        if self._is_status_changed_into_required(changelog_entry):
            if self.interval_start_time is None:
                self.interval_start_time = change_time
        elif self._is_status_changed_from_required(changelog_entry):
            if self.interval_end_time is None:
                self.interval_end_time = change_time
        self._sum_working_time(working_time_per_user, last_assigned_user)

    def _get_current_assignee_from_changelog_when_last_assigned_is_unknown(self, changelog_entry, last_assigned_user):
        if last_assigned_user == self._default_assigned_user():
            status_change_author = self._extract_author_from_changelog_entry(changelog_entry)
            if status_change_author and self._is_allowed_user(status_change_author):
                last_assigned_user = status_change_author
        return last_assigned_user

    @abstractmethod
    def _extract_chronological_changes_sequence(self, task) -> Iterable[dict]:
        pass

    @abstractmethod
    def _is_user_change_entry(self, changelog_entry) -> bool:
        pass

    @abstractmethod
    def _is_status_change_entry(self, changelog_entry) -> bool:
        pass

    @abstractmethod
    def _extract_user_from_change(self, changelog_entry) -> str:
        pass

    @abstractmethod
    def _extract_change_time(self, changelog_entry) -> datetime:
        pass

    @abstractmethod
    def _is_status_changed_into_required(self, changelog_entry) -> bool:
        pass

    @abstractmethod
    def _is_status_changed_from_required(self, changelog_entry) -> bool:
        pass

    @abstractmethod
    def _is_current_status_a_required_status(self, task) -> bool:
        pass

    @abstractmethod
    def _extract_author_from_changelog_entry(self, changelog_entry) -> Optional[str]:
        pass

    @staticmethod
    def _default_assigned_user() -> str:
        return 'UNKNOWN'

    def _is_allowed_user(self, user: Optional[str]) -> bool:
        if self.user_filter is None:
            return True
        if user is None:
            return False
        return user in self.user_filter

    @staticmethod
    def _now() -> datetime:
        try:
            return datetime.now().astimezone()
        except Exception:
            return datetime.now()

    def _sum_working_time(self, working_time_per_user: Dict[str, int], last_assigned_user: str):
        if self._is_interval_found_for_status_change():
            seconds_in_status = self.worktime_extractor.extract_time_from_period(self.interval_start_time,
                                                                                 self.interval_end_time)
            if seconds_in_status is not None:
                working_time_per_user[last_assigned_user] = working_time_per_user.get(last_assigned_user, 0) + seconds_in_status
            self._clean_interval_times()

    def _is_interval_found_for_status_change(self):
        return self.interval_start_time is not None and self.interval_end_time is not None

    def _clean_interval_times(self):
        self.interval_start_time = None
        self.interval_end_time = None
